@@ -1,17 +1,20 @@
 import os
 import json
+import inspect
 from copy import deepcopy
 
 from flask import Blueprint, request, render_template, send_from_directory, current_app
 from flask_restful import (Api as restful_Api, abort as flask_abort,
                            Resource as flask_Resource)
 
+from flask_restful_swagger_3.exceptions import SchemaAlreadyExist
 from flask_restful_swagger_3.swagger import (ValidationError, create_open_api_resource,
                                              add_parameters, validate_path_item_object,
-                                             validate_operation_object,
-                                             validate_components_object,
+                                             validate_components_object, validate_open_api_object,
                                              extract_swagger_path, _auth as auth,
-                                             slash_join, TypeSwagger, REGISTRY_SCHEMA)
+                                             slash_join, REGISTRY_SCHEMA)
+
+from flask_restful_swagger_3.constants import TypeSwagger
 
 from flask_restful_swagger_3.swagger_format import get_validate_format
 
@@ -63,7 +66,11 @@ class Api(restful_Api):
 
         swagger_prefix_url = kwargs.pop("swagger_prefix_url", "/api/doc")
         swagger_url = kwargs.pop("swagger_url", "swagger.json")
-        add_api_spec_resource = kwargs.pop('add_api_spec_resource', True)
+        add_api_spec_resource = kwargs.pop("add_api_spec_resource", True)
+        authorizations = kwargs.pop("authorizations", None)
+
+        if authorizations:
+            self.open_api_object["components"]["securitySchemes"] = authorizations
 
         add_parameters(self.__open_api_object, kwargs)
 
@@ -86,7 +93,7 @@ class Api(restful_Api):
         urls = {}
 
         for method in [m.lower() for m in resource.methods]:
-            __method = {method: {}}
+            operations_object = {method: {}}
             f = resource.__dict__.get(method, None)
             if f:
                 response_code_list = f.__dict__.get("__response_code", [])
@@ -97,16 +104,20 @@ class Api(restful_Api):
                 reqparser = f.__dict__.get("__reqparser", [])
                 tags = f.__dict__.get("__tags", [])
                 no_content_list = f.__dict__.get("__no_content", [])
+                custom_example_list = f.__dict__.get("__custom_example", [])
+                summary_list = f.__dict__.get("__summary", [])
+                security = f.__dict__.get("__security",  [])
 
                 assert (
-                    len(response_code_list) == len(description_list) == len(model_list) == len(no_content_list)
+                    len(response_code_list) == len(description_list) == len(model_list) ==
+                    len(no_content_list) == len(custom_example_list) == len(summary_list)
                 )
 
-                if reqparser and request_body:
-                    raise ValidationError("requestBody and reqparser can't be in same spec")
-
                 if reqparser:
-                    request_body, _params = RequestParserExtractor(reqparser).extract()
+                    parser_json_result, _params = RequestParserExtractor(reqparser).extract()
+                    if parser_json_result and request_body:
+                        raise ValidationError("requestBody and reqparser can't be in same spec")
+                    request_body = parser_json_result
                     params += _params
 
                 result_model = [self.__build_model(model) for model in model_list]
@@ -118,13 +129,13 @@ class Api(restful_Api):
 
                 for result in result_model:
                     if result:
-                        schemas.update(result["schema"])
+                        schemas.update(result["reusable_schema"])
 
                 req_ref = None
                 req_example = None
                 if request_body:
                     req_schema, req_body = self.__build_request_body(request_body)
-                    __method[method].update(req_body)
+                    operations_object[method].update(req_body)
                     if req_schema:
                         schemas.update(req_schema)
 
@@ -134,12 +145,19 @@ class Api(restful_Api):
                             if req_result_model
                             else None
                         )
+
                     req_example = (
                         req_result_model["example"]
                         if req_result_model
                         else None
                     )
 
+                if security:
+                    if "security" in operations_object:
+                        operations_object[method]["security"].append(security)
+                    else:
+                        operations_object[method]["security"] = [security]
+
                 for index, response_code in enumerate(response_code_list):
                     ref = (
                         result_model[index]["reference"]
@@ -147,52 +165,69 @@ class Api(restful_Api):
                         else None
                     )
                     example_schema = (
-                        result_model[index]["example"] if result_model[index] else None
+                        result_model[index]["example"]
+                        if result_model[index]
+                        else None
                     )
                     response = self.__build_responses(
                         response_code,
                         ref=ref or req_ref,
-                        example_schema=example_schema or req_example,
+                        custom_example=custom_example_list[index],
+                        example=example_schema or req_example,
                         description=description_list[index],
                         no_content=no_content_list[index]
                     )
 
                     for url in args:
-                        if not url.startswith('/'):
-                            raise ValidationError('paths must start with a /')
+                        if url.endswith('/'):
+                            raise ValidationError('paths must not have ending slash')
                         if self.blueprint and self.blueprint.url_prefix:
                             if not self.blueprint.url_prefix.startswith('/'):
-                                raise ValidationError('url_prefix must start with a /')
+                                raise ValidationError('url_prefix must start with a leading slash')
                             if self.blueprint.url_prefix.endswith('/'):
-                                raise ValidationError('url_prefix must not end with a /')
+                                raise ValidationError('url_prefix must not have ending slash')
                             url = self.blueprint.url_prefix + url
 
                         converted_url, parameters = self.__build_parameters(url, params)
 
-                        __method[method]['tags'] = tags
-                        __method[method].update(parameters)
+                        operations_object[method]['tags'] = tags
+                        operations_object[method].update(parameters)
 
-                        if "responses" in __method[method]:
-                            __method[method]["responses"].update(response)
+                        if "responses" in operations_object[method]:
+                            operations_object[method]["responses"].update(response)
                         else:
-                            __method[method]["responses"] = response
+                            operations_object[method]["responses"] = response
 
-                        validate_path_item_object(__method)
+                        if "summary" not in operations_object[method]:
+                            if summary_list[index]:
+                                operations_object[method]["summary"] = summary_list[index]
+
+                            elif len(tags) > 0:
+                                operations_object[method]["summary"] = f"Operations on {', '.join(tags).lower()}"
+
+                        if "securitySchemes" in self.open_api_object["components"]:
+                            components_security_schemes = self.open_api_object["components"]["securitySchemes"]
+                        else:
+                            components_security_schemes = None
+                        validate_path_item_object(operations_object, components_security_schemes)
 
                         if converted_url in urls:
-                            urls[converted_url].update(__method)
+                            urls[converted_url].update(operations_object)
                         else:
-                            urls[converted_url] = __method
+                            urls[converted_url] = operations_object
 
         self.__open_api_object["paths"].update(urls)
 
-        validate_components_object(schemas)
-
         if "schemas" in self.__open_api_object["components"]:
             self.__open_api_object["components"]["schemas"].update(schemas)
         else:
             self.__open_api_object["components"]["schemas"] = schemas
 
+        if 'externalDocs' in self.__open_api_object and not self.__open_api_object['externalDocs']:
+            del self.__open_api_object['externalDocs']
+
+        validate_open_api_object(self.open_api_object)
+
         super().add_resource(resource, *args, endpoint=endpoint, **kwargs)
 
     @staticmethod
@@ -205,26 +240,37 @@ class Api(restful_Api):
     def __build_model(schema):
         if schema:
             is_list = type(schema) == list
-            schema_name = schema[0].__name__ if is_list else schema.__name__
+            try:
+                schema_name = schema[0].__name__ if is_list else schema.__name__
+                if schema_name not in REGISTRY_SCHEMA:
+                    raise TypeError("'schema' used with 'reorder_with' must be a sub class of Schema")
+            except AttributeError:
+                raise TypeError("'schema' used with 'reorder_with' must be a sub class of Schema")
 
             definition = schema[0].definitions() if is_list else schema.definitions()
             reference = schema[0].reference() if is_list else schema.reference()
-            _schema = eval(json.dumps(definition, cls=DefinitionEncoder))
+            _schema = json.loads(json.dumps(definition, cls=DefinitionEncoder))
 
             example = (
-                [eval(json.dumps(schema[0].example(), cls=ExampleEncoder))]
+                [json.loads(json.dumps(schema[0].example(), cls=ExampleEncoder))]
                 if is_list
-                else eval(json.dumps(schema.example(), cls=ExampleEncoder))
+                else json.loads(json.dumps(schema.example(), cls=ExampleEncoder))
             )
 
+            schema_example_name = schema[0].reference_example_name() if is_list else schema.reference_example_name()
+            reference_example = schema[0].reference_example() if is_list else schema.reference_example()
+
             return {
-                "schema": {schema_name: _schema},
+                "reusable_schema": {schema_name: _schema},
                 "reference": reference,
-                "example": example,
+                "reference_example": {schema_example_name: reference_example},
+                "reusable_example": {schema_example_name: {'value': example}},
+                "example": example
             }
 
     @staticmethod
-    def __build_responses(response_code, description="", ref=None, example_schema=None, no_content=False):
+    def __build_responses(response_code, description="", ref=None,
+                          custom_example=None, example=None, no_content=False):
         responses = {response_code: {"content": {"application/json": {}}}}
 
         if description:
@@ -234,8 +280,12 @@ class Api(restful_Api):
             _schema = {"schema": ref}
             responses[response_code]["content"]["application/json"].update(_schema)
 
-        if example_schema:
-            _example = {"example": example_schema}
+        if custom_example:
+            custom_example = {"example": custom_example}
+            responses[response_code]["content"]["application/json"].update(custom_example)
+
+        if example:
+            _example = {"example": example}
             responses[response_code]["content"]["application/json"].update(_example)
 
         if no_content:
@@ -251,7 +301,8 @@ class Api(restful_Api):
             converter_variable = param.split(":")
             if len(converter_variable) > 2:
                 raise ValueError(
-                    f"You must define one converter for a variable_name, if you want several converter don't mention any '{param}'"
+                    f"You must define one converter for a variable_name, "
+                    f"if you want several converter don't mention any '{':'.join(converter_variable[:-1])}'"
                 )
 
             try:
@@ -306,18 +357,14 @@ class Api(restful_Api):
             model = self.__build_model(schema)
             reference = {"schema": model["reference"]}
 
-            result["requestBody"]["content"]["application/json"] = reference
+            result["requestBody"]["content"]["application/json"].update(reference)
 
-            return model["schema"], result
+            return model["reusable_schema"],  result
 
         return None, result
 
     @property
-    def blueprint_name(self):
-        return repr(super().__class__)
-
-    @property
-    def open_api_json(self):
+    def open_api_object(self):
         return self.__open_api_object
 
 
@@ -340,7 +387,6 @@ class RequestParserExtractor:
         if "parser" not in reqparser:
             raise ValidationError("parser must be define in reqparser")
         return self._get_reqparse_args(reqparser)
-        # return self._extract_schemas(operation)
 
     def _get_reqparse_args(self, reqparser):
         """
@@ -386,8 +432,6 @@ class RequestParserExtractor:
             return type_.swagger_type
         elif callable(type_) and type_.__name__ == 'boolean':  # flask-restful boolean
             return 'boolean'
-        elif issubclass(type_, str):
-            return 'string'
         elif type_ == float:
             return 'float'
         elif type_ == int:
@@ -401,9 +445,9 @@ class RequestParserExtractor:
         elif type_ == dict:
             return 'object'
         try:
-            if type_ == long:
-                return 'long'
-        except NameError:
+            if issubclass(type_, str):
+                return 'string'
+        except TypeError:
             pass
         raise TypeError('unexpected type: {0}'.format(type_))
 
@@ -457,6 +501,8 @@ class RequestParserExtractor:
 
 
 def register_schema(target_class):
+    if target_class.__name__ in REGISTRY_SCHEMA:
+        raise SchemaAlreadyExist(target_class.__name__)
     REGISTRY_SCHEMA[target_class.__name__] = target_class
 
 
@@ -464,36 +510,41 @@ class Schema(dict):
     properties = None
 
     def __init_subclass__(cls, **kwargs):
-        if cls not in REGISTRY_SCHEMA:
-            register_schema(cls)
+        register_schema(cls)
         super().__init_subclass__(**kwargs)
-        super_classes = cls.get_super_classes(cls)
+        super_classes = cls.get_super_classes()
         properties = {}
         if cls.properties:
-            properties.update(deepcopy(cls.properties))
+            if type(cls.properties) is not dict:
+                raise TypeError(f"Attribute 'properties' must be a 'dict', but was {type(cls.properties)}")
+            properties = cls.update_properties(properties, cls.properties)
+        if hasattr(cls, 'required'):
+            if type(cls.required) not in [list, set, tuple]:
+                raise TypeError(f"Attribute 'required' must be 'list', 'set' or 'tuple', but was {type(cls.required)}")
         for super_class in super_classes:
-            if not hasattr(super_class, 'type'):
-                raise TypeError("You can inherit only schema of type 'object'")
             if super_class.type != 'object':
                 raise TypeError("You can inherit only schema of type 'object'")
             if cls.type != super_class.type:
-                raise TypeError(f"You can't add type to '{cls.__name__}'" +
+                raise TypeError(f"You can't add type to '{cls.__name__}' " +
                                 f"because it inherits of type of '{super_class.__name__}'")
 
             cls.type = super_class.type
 
             if super_class.properties:
-                properties.update(deepcopy(super_class.properties))
+                properties = cls.update_properties(super_class.properties, properties)
 
             if hasattr(super_class, 'required'):
                 if hasattr(cls, 'required'):
                     cls.required = list(set(cls.required + super_class.required))
-                else:
-                    cls.required = super_class.required
 
         if properties:
             cls.properties = properties
 
+        cls.check_schema_type()
+
+        if cls.properties and not hasattr(cls, 'type'):
+            cls.type = 'object'
+
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
 
@@ -505,46 +556,112 @@ class Schema(dict):
                 if type(self.properties[k]) == type:
                     if self.properties[k].type == 'object':
                         self.properties[k](**v)
-                    prop = self.properties[k].definitions()
+                    self.prop = self.properties[k].definitions()
                 else:
-                    prop = self.properties[k]
+                    self.prop = self.properties[k]
 
-                nullable = False
-                if 'nullable' in prop:
-                    if prop['nullable'] not in ['true', 'false']:
-                        raise ValueError('\'nullable\' must be \'true\' or \'false\'')
-                    if prop['nullable'] == 'true':
-                        nullable = True
+                nullable = self.get_boolean_attribute('nullable')
+                load_only = self.get_boolean_attribute('load_only')
+                dump_only = self.get_boolean_attribute('dump_only')
+                if load_only and dump_only:
+                    raise TypeError('A value can\'t be load_only and dump_only in the same schema')
 
-                if nullable and v is None:
+                type_ = self.prop.get('type', None)
+                format_ = self.prop.get('format', None)
+
+                if not (nullable and v is None):
+                    self.check_type(type_, k, v)
+                    if 'enum' in self.prop:
+                        if type(self.prop['enum']) not in [set, list, tuple]:
+                            raise TypeError(f"'enum' must be 'list', 'set' or 'tuple',"
+                                            f"but was {type(self.prop['enum'])}")
+                        for item in list(self.prop['enum']):
+                            self.check_type(type_, 'enum', item)
+                        if v not in self.prop['enum']:
+                            raise ValueError(f"{k} must have {' or '.join(self.prop['enum'])} but have {v}")
+
+                    self.check_format(type_, format_, v)
+
+                if load_only:
+                    del self[k]
                     continue
 
-                type_ = prop.get('type', None)
-                format_ = prop.get('format', None)
-                self.check_type(type_, format_, k, v)
-
-                if 'enum' in prop:
-                    if type(prop['enum']) not in [set, list, tuple]:
-                        raise TypeError(f"'enum' is must be 'list', 'set' or 'tuple', but was {type(prop['enum'])}")
-                    for item in list(prop['enum']):
-                        self.check_enum_type(type_, format_, item)
-                    if v not in prop['enum']:
-                        raise ValueError(f"{k} must have {' or '.join(prop['enum'])} but have {v}")
                 self[k] = v
 
         if hasattr(self, 'required'):
+            self.required = list(self.required)
             for key in self.required:
                 if key not in kwargs:
                     raise ValueError('The attribute "{0}" is required'.format(key))
 
+    def get_boolean_attribute(self, attr):
+        _attr = False
+        if attr in self.prop:
+            if self.prop[attr] not in ['true', 'false', True, False]:
+                raise ValueError(f'"{attr}" must be "true", "false", True, False')
+            if self.prop[attr] == 'true' or self.prop[attr]:
+                _attr = True
+
+        return _attr
+
+    @classmethod
+    def check_schema_type(cls):
+        if hasattr(cls, 'type'):
+            if cls.type == 'object' and not cls.properties:
+                raise TypeError("Attribute properties cannot be None when schema type is object")
+            if cls.type == 'array':
+                if (not type(cls.items) is dict) or (inspect.isclass(cls.items) and not issubclass(cls.items, Schema)):
+                    raise TypeError(f"Attribute items must be of type dict or a subclass of Schema,"
+                                    f" but was {type(cls.items)}")
+
     @staticmethod
+    def update_properties(original, new):
+        if not original:
+            return new
+        updated = deepcopy(original)
+        try:
+            new_ = new.definitions()
+        except AttributeError:
+            new_ = new
+        for prop, dict_or_schema in new_.items():
+            try:
+                new_prop = new_[prop].definitions()
+            except AttributeError:
+                new_prop = new_[prop]
+            if prop in updated:
+                try:
+                    _dict_or_schema = dict_or_schema.definitions()
+                except AttributeError:
+                    _dict_or_schema = dict_or_schema
+                for k, v in _dict_or_schema.items():
+                    try:
+                        updated_prop = updated[prop].definitions()
+                    except AttributeError:
+                        updated_prop = updated[prop]
+                    if k in updated_prop.keys():
+                        if k == 'type' and updated_prop[k] != v:
+                            raise TypeError("You can't alter type of properties in sub schema")
+
+                try:
+                    if updated[prop].type == 'object':
+                        updated[prop] = updated[prop].update_properties(updated[prop].properties, new[prop].properties)
+                    else:
+                        updated[prop] = new_prop
+                except AttributeError:
+                    updated[prop].update(new_prop)
+            else:
+                updated[prop] = new_prop
+
+        return updated
+
+    @classmethod
     def get_super_classes(cls):
         return [
             schema for schema_name, schema in REGISTRY_SCHEMA.items()
             if schema_name != cls.__name__ and issubclass(cls, schema)
         ]
 
-    def check_type(self, type_, format_, key, value):
+    def check_type(self, type_, key, value):
         if type_:
             if type_ == 'array':
                 if not isinstance(value, list):
@@ -555,7 +672,8 @@ class Schema(dict):
                         for v in value:
                             cls(**v)
                     else:
-                        self.check_array_type(cls.type, format_,  key, value)
+                        for v in value:
+                            self.check_type(cls.type,  key, v)
             if type_ == 'integer' and not isinstance(value, int):
                 raise ValueError(f'The attribute "{key}" must be an int, but was "{type(value)}"')
             if type_ == 'number' and not isinstance(value, int) and not isinstance(value, float):
@@ -564,34 +682,7 @@ class Schema(dict):
             if type_ == 'string' and not isinstance(value, str):
                 raise ValueError(f'The attribute "{key}" must be a string, but was "{type(value)}"')
             if type_ == 'boolean' and not isinstance(value, bool):
-                raise ValueError(f'The attribute "{key}" must be an int, but was "{type(value)}"')
-            self.check_format(type_, format_, value)
-
-    def check_array_type(self, type_, format_, key, value):
-        if type_:
-            if type_ == 'integer' and not all([isinstance(v, int) for v in value]):
-                raise ValueError(f'The list "{key}" must have all items of type int')
-            if type_ == 'number' and not all([isinstance(v, int) for v in value]) and not all([isinstance(v, float) for v in value]):
-                raise ValueError(
-                    f'The list "{key}" must have all items of type int or float')
-            if type_ == 'string' and not all([isinstance(v, str) for v in value]):
-                raise ValueError(f'The list "{key}" must have all items of type string')
-            if type_ == 'boolean' and not all([isinstance(v, bool) for v in value]):
-                raise ValueError(f'The list "{key}" must have all items of type string')
-            self.check_format(type_, format_, value)
-
-    def check_enum_type(self, type_, format_, value):
-        if type_:
-            if type_ == 'integer' and not isinstance(value, int):
-                raise ValueError(f'The enum "{value}" must be an int, but was "{type(value)}"')
-            if type_ == 'number' and not isinstance(value, int) and not isinstance(value, float):
-                raise ValueError(
-                    f'The enum "{value}" must be an int or float, but was "{type(value)}"')
-            if type_ == 'string' and not isinstance(value, str):
-                raise ValueError(f'The enum "{value}" must be a string, but was "{type(value)}"')
-            if type_ == 'boolean' and not isinstance(value, bool):
-                raise ValueError(f'The enum "{value}" must be an int, but was "{type(value)}"')
-            self.check_format(type_, format_, value)
+                raise ValueError(f'The attribute "{key}" must be a bool, but was "{type(value)}"')
 
     @staticmethod
     def check_format(type_, format_, value):
@@ -601,7 +692,15 @@ class Schema(dict):
 
     @classmethod
     def reference(cls):
-        return {'$ref': '#/components/schemas/{0}'.format(cls.__name__)}
+        return {'$ref': f'#/components/schemas/{cls.__name__}'}
+
+    @classmethod
+    def reference_example_name(cls):
+        return f'Example{cls.__name__}'
+
+    @classmethod
+    def reference_example(cls):
+        return {'$ref': f'#/components/examples/{cls.reference_example_name()}'}
 
     @classmethod
     def definitions(cls):
@@ -618,22 +717,44 @@ class Schema(dict):
     @classmethod
     def example(cls):
         items = dict(cls.__dict__.items())
-        if "properties" in items:
-            properties = dict(cls.__dict__.items())["properties"]
-            example = {}
-            if properties:
-                for k, v in properties.items():
-                    if type(v) is dict:
-                        val = v["type"]
-                        if v["type"] == "array":
-                            if "items" in v:
-                                val = [v["items"].example()]
-                            else:
-                                val = []
-                    else:
-                        val = [] if v == "array" else v
-                    example.update({k: val})
-            return example
+        if "type" in items:
+            if items["type"] == "object":
+                return cls.__example_object(items)
+            else:
+                return cls.__example(items)
+
+    @staticmethod
+    def __example_object(items):
+        properties = items["properties"]
+        example = {}
+        for k, v in properties.items():
+            try:
+                load_only = 'load_only' in v and v['load_only']
+            except TypeError:
+                load_only = hasattr(v, 'load_only') and v.load_only
+            try:
+                if load_only:
+                    continue
+                val = v["type"]
+                if v["type"] == "array":
+                    if "items" in v:
+                        try:
+                            val = [v["items"].example()]
+                        except AttributeError:
+                            val = [v["items"]["type"]]
+
+            except TypeError:
+                val = [] if v == "array" else v
+            example.update({k: val})
+        return example
+
+    @staticmethod
+    def __example(items):
+        if items["type"] == 'array':
+            try:
+                return [items.example()]
+            except AttributeError:
+                return [items['items']["type"]]
         return items["type"]
 
 
@@ -665,6 +786,8 @@ def get_swagger_blueprint(
     """
 
     add_parameters(swagger_object, kwargs)
+    validate_open_api_object(swagger_object)
+
     app_name = kwargs.get('title', 'Swagger UI')
     swagger_blueprint_name = kwargs.get('swagger_blueprint_name', 'swagger')
 
